import numpy as np
from sklearn import preprocessing, metrics
import scipy.stats
import pickle
import warnings
import src.atmos_physics as atmos_physics
from netCDF4 import Dataset


def LoadData(filename, min_lev, all_lats=True, ind_lat=None, n_trn_exs=None,
             rain_only=False, no_cos=False, verbose=False, use_rh=False):
    """v2 of the script to load data. See prep_convection_output.py for how
       the input filename is generated.

    Args:
      filename:  The file to be loaded. e.g., './data/conv_cond_training_v3.pkl'
      min_lev:    The topmost model level for which to load data. Set to 0. to
                 load all data
      all_lats:  Logical value for whether to load data from all latitudes
      ind_lat:    If all_lats is false, give the index value [0-63] for the
                 latitude at which to load data.
      n_trn_exs: Number of training examples to load. If set to None, or
                 if requested number exceeds max available will load all.
      rain_only:  If true, only return training examples of when it is raining
      no_cos:   If true, don't use cos(lat) weighting for loading training examples
      verbose:   If true, prints some basic stats about training set

    Returns:
      x       : 2-d numpy array of input features (m_training examples x
                n_input features). If min_lev is 0., there will be 61 input
                features, 30 for temperature, 30 for humidity, 
                1 for surface pressure
      y       : 2-d numpy array of output targets (m_traning examples x
                n_output targets). If min_lev is 0., there will be 60 output
                features, the top 30 for temp. tendencies and the bottom 30
                for q tend.
      cv      : 1-d array (m_training examples x 1) that gives 1 if convection
                occurs and 0 if it does not.
      Pout    : 1-d arrray (m_training examples x 1) of how much precipitation
                occurs in kg/m^2/s 
      lev     : The vertical model levels (1 is the surface and 0 is the top
                of the atmosphere).
      dlev    : The difference between model levels, useful for calculating
                some derived quantities.
    """
    # Data to read in is n_lev x n_lat (SH & NH) x n_samples
    # Samples are quasi indpendent with only a few (e.g. 10) from each latitude range and sampled once a day
    v = dict()
    [v['Tin'], v['qin'], psin, v['Tout'], v['qout'], Pout, lat, half_lev] = \
        pickle.load(open(filename, 'rb'))

    # Calculate the sigma levels
    lev, dlev, indlev = get_levs(min_lev, half_lev)

    varis = ['Tin', 'qin', 'Tout', 'qout']
    # Reshape the arrays
    for var in varis:
        # Change shape of data to be n_samp x n_lev
        if all_lats:
            if no_cos:
                v[var] = reshape_all_lats(v[var], indlev)
            else:
                v[var] = reshape_cos_lats(v[var], indlev, lat)
        else:
            if ind_lat is not None:
                v[var] = reshape_one_lat(v[var], indlev, ind_lat)
            else:
                raise TypeError('Need to set an index value for ind_lat')
    # Also reshape precipitation and surface pressure
    if all_lats:
        if no_cos:
            # Need to do a transpose to be consistent with reshape_all_lats
            psin = np.reshape(psin.transpose(), -1)
            Pout = np.reshape(Pout.transpose(), -1)
        else:
            psin = reshape_cos_lats(psin, None, lat, is_sfc=True)
            Pout = reshape_cos_lats(Pout, None, lat, is_sfc=True)
    else:
        psin = psin[ind_lat, :]
        Pout = Pout[ind_lat, :]

    # Use relative humidity instead of specific humidity as a feature
    if use_rh:
      esat = atmos_physics.simple_sat_vap(v['Tin'])
      p = np.zeros(v['qin'].shape)
      for k in range(v['qin'].shape[1]):
       p[:,k] = psin*lev[k]
      e = atmos_physics.vap(p, v['qin'])
      v['qin'] = e/esat

     
    # Randomize the order of these events to ensure different latitudes are mixed up
    np.random.seed(123) # seed random number generator so always get same set of data when call this function (e.g., if call again to make plots)
    m = v['Tin'].shape[0]
    randind = np.random.permutation(m)
    for var in varis:
        v[var] = v[var][randind, :]
    Pout = Pout[randind]
    psin = psin[randind]


    # Concatenate input and output variables together
    x = pack_x(v['Tin'], v['qin'], psin)
    y = pack_y(v['Tout'], v['qout'], axis=1)

    # Print some statistics about rain and limit to when it's raining if True
    x, y, Pout = limitrain(x, y, Pout, rain_only, verbose=verbose)

    # Limit to only certain events if requested
    if n_trn_exs is not None:
        if n_trn_exs > y.shape[0]:
            warnings.warn('Requested more samples than available. Using the' +
                          'maximum number available')
            n_trn_exs = y.shape[0]
        ind = np.arange(n_trn_exs)
        x = x[ind, :]
        y = y[ind, :]
        Pout = Pout[ind]

    # Store when convection occurs
    cv, _ = whenconvection(y, verbose=verbose)

    return x, y, cv, Pout, lat, lev, dlev


def reshape_cos_lats(z, indlev, lat, is_sfc=False):
    if is_sfc:
        z = z.swapaxes(0, 1)
        z2 = np.empty((0))
    else:
        z = z[indlev, :, :]
        z = z.swapaxes(0, 2)
        z2 = np.empty((0, sum(indlev)))
    n_ex = z.shape[0]
    for i, latval in enumerate(lat):
        Ninds = int(n_ex * np.cos(np.deg2rad(latval)))
        # xtropics or tropics (also change ml_train.py)
        #Ninds = int(n_ex * np.cos(np.deg2rad(latval))*(np.abs(latval)>20.0))
        #Ninds = int(n_ex * np.cos(np.deg2rad(latval))*(np.abs(latval)<=20.0))
        if is_sfc:
            z2 = np.concatenate((z2, z[0: Ninds, i]), axis=0)
        else:
            z2 = np.concatenate((z2, z[0:Ninds, i, :]), axis=0)
    return z2


def reshape_all_lats(z, indlev):
    # Expects data to be n_lev x n_lat x n_samples and returns
    # (n_lat*n_samp x n_lev)
    z = z[indlev, :, :]
    z = z.swapaxes(0, 2)
    return np.reshape(z, (-1, sum(indlev)))


def reshape_one_lat(z, indlev, ind_lat):
    # Expects data to be n_lev x n_lat x n_samples and returns (n_samp x n_lev)
    z = z[indlev, ind_lat, :]
    z = z.swapaxes(0, 1)
    return z

def pack_x(T, q, ps):
    """Combines T & q profiles and ps"""
    return np.concatenate((T, q, ps[:,np.newaxis]), axis=1)

def unpack_x(data, vari, axis=1):
    """Reverse pack operation"""
    N = int((data.shape[axis]-1)/2)
    varipos = {'T': np.arange(N), 'q': np.arange(N, 2*N), 'ps': 2*N}
    out = np.take(data, varipos[vari], axis=axis)
    return out

def pack_y(d1, d2, axis=1):
    """Combines T & q profiles"""
    return np.concatenate((d1, d2), axis=axis)


def unpack_y(data, vari, axis=1):
    """Reverse pack operation"""
    N = int(data.shape[axis]/2)
    varipos = {'T': np.arange(N), 'q': np.arange(N, 2*N)}
    out = np.take(data, varipos[vari], axis=axis)
    return out


# Initialize & fit scaler
def init_pp(ppi, raw_data):
    # Initialize list of scaler objects
    if ppi['name'] == 'MinMax':
        pp = preprocessing.MinMaxScaler(feature_range=(-1.0, 1.0))
        pp.fit(raw_data)
    elif ppi['name'] == 'MaxAbs':
        pp = preprocessing.MaxAbsScaler() 
        pp.fit(raw_data)
    elif ppi['name'] == 'StandardScaler':
        pp = preprocessing.StandardScaler() 
        pp.fit(raw_data)
    elif ppi['name'] == 'RobustScaler':
        pp = preprocessing.RobustScaler()
        pp.fit(raw_data)
    elif ppi['name'] == 'SimpleY' or ppi['name'] == 'SigmaCubedY' or ppi['name'] == 'SigmaSquaredY':
        pp = [atmos_physics.cp, atmos_physics.L]  
    elif ppi['name'] == 'NoScaler':
        pp = []
    else:
        raise ValueError('Incorrect scaler name')

    return pp


# Transform data using initialized scaler
def transform_data(ppi, pp, raw_data, lev):
    if ppi['name'] == 'SimpleY':
        T_data = unpack_y(raw_data, 'T')*pp[0]
        q_data = unpack_y(raw_data, 'q')*pp[1]
        return_data = pack_y(T_data, q_data)
    elif ppi['name'] == 'SigmaCubedY':
        T_data = unpack_y(raw_data, 'T')*pp[0]
        # weight upper levels to a greater extent for humidity
        q_data = unpack_y(raw_data, 'q')*pp[1]/(lev**3)
        return_data = pack_y(T_data, q_data)
    elif ppi['name'] == 'SigmaSquaredY':
        T_data = unpack_y(raw_data, 'T')*pp[0]
        # weight upper levels to a greater extent for humidity
        q_data = unpack_y(raw_data, 'q')*pp[1]/(lev**2)
        return_data = pack_y(T_data, q_data)
    elif ppi['name'] == 'NoScaler':
        return_data = raw_data
    else:
        return_data = pp.transform(raw_data)

    # Return single transformed array as output
    return return_data 


# Apply inverse transformation to unscale data
def inverse_transform_data(ppi, pp, trans_data, lev):
    if ppi['name'] == 'SimpleY':
        T_data = unpack_y(trans_data, 'T')/pp[0]
        q_data = unpack_y(trans_data, 'q')/pp[1]
        return_data = pack_y(T_data, q_data)
    elif ppi['name'] == 'SigmaCubedY':
        T_data = unpack_y(trans_data, 'T')/pp[0]
        q_data = unpack_y(trans_data, 'q')/pp[1]*(lev**3)
        return_data = pack_y(T_data, q_data)
    elif ppi['name'] == 'SigmaSquaredY':
        T_data = unpack_y(trans_data, 'T')/pp[0]
        q_data = unpack_y(trans_data, 'q')/pp[1]*(lev**2)
        return_data = pack_y(T_data, q_data)
    elif ppi['name'] == 'NoScaler':
        return_data = trans_data
    else:
        return_data = pp.inverse_transform(trans_data)
    return return_data


def limitrain(x, y, Pout, rain_only=False, verbose=True):
    indrain = np.greater(Pout, 0)
    if verbose:
        print('There is some amount of rain {:.1f}% of the time'.
              format(100. * np.sum(indrain)/len(indrain)))
    if rain_only:
        x = x[indrain, :]
        y = y[indrain, :]
        Pout = Pout[indrain]
        if verbose:
            print('Only looking at times it is raining!')
    return x, y, Pout


def whenconvection(y, verbose=True):
    """Caluclate how often convection occurs...useful for classification
       Also store a variable that is 1 if convection and 0 if no convection"""
    cv_strength = np.sum(np.abs(unpack_y(y, 'T')), axis=1)
    cv = np.copy(cv_strength)
    cv[cv > 0] = 1
    if verbose:
        print('There is convection (or condensation if conv_cond) {:.1f}% of the time'.
              format(100. * np.sum(cv)/len(cv)))
    return cv, cv_strength


def avg_hem(data, lat, axis, split=False):
    """Averages the NH and SH data (or splits them into two data sets)"""
    ixsh = np.where(lat < 0)[0]  # where returns a tuple
    ixnh = np.where(lat >= 0)[0]
    if len(ixsh) == 0:
        print(lat)
        raise ValueError('Appears that lat does not have SH values')
    lathalf = lat[ixnh]
    sh = np.take(data, ixsh, axis=axis)
    nh = np.take(data, ixnh, axis=axis)
    # Flip the direction of the sh data at a given axis
    shrev = np.swapaxes(np.swapaxes(sh, 0, axis)[::-1], 0, axis)
    # If splitting data, return these arrays
    if split:
        return nh, shrev, lathalf
    else:
        return (nh + shrev) / 2., lathalf


def load_one_lat(x_ppi, y_ppi, x_pp, y_pp, est, ind_lat, datafile, min_lev,
                 n_trn_exs, rain_only, no_cos, no_ps, use_rh):
    """Returns n_samples x 2*n_lev array of true and predicted values
       at a given latitude"""
    # Load data
    x, y, cv, Pout, lat, lev, dlev = \
        LoadData(datafile, min_lev, all_lats=False, ind_lat=ind_lat, 
                 verbose=False, n_trn_exs=None, rain_only=rain_only, 
                 no_cos=no_cos, use_rh=use_rh)
    # Calculate predicted output
    x_scl = transform_data(x_ppi, x_pp, x, lev)

    if no_ps: # set ps_scl to a constant for training
      T_scl = unpack_x(x_scl, 'T')
      q_scl = unpack_x(x_scl, 'q')
      ps_scl = unpack_x(x_scl, 'ps')
      y_pred_scl = est.predict(pack_x(T_scl,q_scl,np.zeros(ps_scl.shape)))
    else:
     y_pred_scl = est.predict(x_scl)

    y_pred = inverse_transform_data(y_ppi, y_pp, y_pred_scl, lev)
    # Output true and predicted temperature and humidity tendencies
    T = unpack_y(y, 'T')
    q = unpack_y(y, 'q')
    T_pred = unpack_y(y_pred, 'T')
    q_pred = unpack_y(y_pred, 'q')
    return T, q, T_pred, q_pred


def stats_by_latlev(x_ppi, y_ppi, x_pp, y_pp, est, lat, lev, datafile, n_trn_exs, rain_only, no_cos, no_ps, use_rh):
    # Initialize
    Tmean = np.zeros((len(lat), len(lev)))
    qmean = np.zeros((len(lat), len(lev)))
    Tvar = np.zeros((len(lat), len(lev)))
    qvar = np.zeros((len(lat), len(lev)))
    Tbias = np.zeros((len(lat), len(lev)))
    qbias = np.zeros((len(lat), len(lev)))
    rmseT = np.zeros((len(lat), len(lev)))
    rmseq = np.zeros((len(lat), len(lev)))
    rT = np.zeros((len(lat), len(lev)))
    rq = np.zeros((len(lat), len(lev)))
    Rsq_T = np.zeros((len(lat), len(lev)))
    Rsq_q = np.zeros((len(lat), len(lev)))
    for i in range(len(lat)):
        #print('Loading data for latitude {:d} of {:d}'.format(i, len(lat)))
        T_true, q_true, T_pred, q_pred = \
            load_one_lat(x_ppi, y_ppi, x_pp, y_pp, est, i, datafile,
                         np.min(lev), n_trn_exs, rain_only,  
                         no_cos, no_ps, use_rh)
        if i==0:
         print('size of test dataset for a given latitude and level', T_true.shape[0])
        # Get mean and variance of true output
        Tmean[i, :] = np.mean(T_true, axis=0)
        qmean[i, :] = np.mean(q_true, axis=0)
        Tvar[i, :] = np.var(T_true, axis=0)
        qvar[i, :] = np.var(q_true, axis=0)
        # Get bias from means
        Tbias[i, :] = np.mean(T_pred, axis=0) - Tmean[i, :]
        qbias[i, :] = np.mean(q_pred, axis=0) - qmean[i, :]
        # Get rmse
        rmseT[i, :] = np.sqrt(
            metrics.mean_squared_error(T_true, T_pred,
                                       multioutput='raw_values'))
        rmseq[i, :] = np.sqrt(
            metrics.mean_squared_error(q_true, q_pred,
                                       multioutput='raw_values'))
        # Get correlation coefficients
        for j in range(len(lev)):
            rT[i, j], _ = scipy.stats.pearsonr(T_true[:, j], T_pred[:, j])
            rq[i, j], _ = scipy.stats.pearsonr(q_true[:, j], q_pred[:, j])

        # Get coefficient of determination
        for j in range(len(lev)):
            Rsq_T[i, j] = metrics.r2_score(T_true[:, j], T_pred[:, j])
            Rsq_q[i, j] = metrics.r2_score(q_true[:, j], q_pred[:, j])

    return Tmean.T, qmean.T, Tvar.T, qvar.T, Tbias.T, qbias.T, rmseT.T, rmseq.T, rT.T, rq.T, Rsq_T.T, Rsq_q.T


def GetDataPath(conv_cond, training_expt):
    datadir = '/net/esker/archive1/pog/mldata/training_data/'
    if conv_cond:
        trainfile = datadir + training_expt + '_conv_cond_training.pkl'
        testfile = datadir + training_expt + '_conv_cond_testing.pkl'
        pp_str = 'conv_cond_' + training_expt + '_'
    else:
        trainfile = datadir + training_expt + '_conv_training.pkl'
        testfile = datadir + training_expt + '_conv_testing.pkl'
        pp_str = training_expt + '_'
 
    print(trainfile)
    print(testfile)
    return datadir, trainfile, testfile, pp_str


def get_levs(min_lev, half_lev):
    # Calculate the full levels
    lev = np.array(np.zeros((half_lev.size-1,)))
    for i in range(half_lev.size-1):
        lev[i] = (half_lev[i] + half_lev[i+1])/2.
    # Limit levels to those specified
    indlev = np.greater_equal(lev, min_lev)
    lev = lev[indlev]
    # Calculate the spacing between levels
    dlev = np.diff(half_lev)
    dlev = dlev[indlev]
    return lev, dlev, indlev

def get_x_y_pred_true(est_str, training_file, min_lev, 
                      all_lats=True, ind_lat=None, 
                      n_trn_exs=None, rain_only=False,  
                      no_cos=False, no_ps=False, use_rh=False):
    # Load model and preprocessors
    est, _, errors, x_ppi, y_ppi, x_pp, y_pp, lat, lev, _ = \
        pickle.load(open('/net/esker/archive1/pog/mldata/regressors/' + est_str + '.pkl', 'rb'))
    # Load raw data from file
    x, ytrue, _, Pout, _, _, _ = \
        LoadData(training_file, min_lev=min_lev, all_lats=all_lats, ind_lat=ind_lat, n_trn_exs=n_trn_exs, rain_only=rain_only, no_cos=no_cos, use_rh=use_rh)
    # Scale true values
    ytrue_scl = transform_data(y_ppi, y_pp, ytrue, lev)

    # Apply x preprocessing to scale x-data and predict output
    x_scl = transform_data(x_ppi, x_pp, x, lev)

    if no_ps: # set ps_scl to a constant for training
      T_scl = unpack_x(x_scl, 'T')
      q_scl = unpack_x(x_scl, 'q')
      ps_scl = unpack_x(x_scl, 'ps')
      ypred_scl = est.predict(pack_x(T_scl,q_scl,np.zeros(ps_scl.shape)))
    else:
     ypred_scl = est.predict(x_scl)

    ypred = inverse_transform_data(y_ppi, y_pp, ypred_scl, lev)
    return x_scl, ypred_scl, ytrue_scl, x, ypred, ytrue, Pout


def load_error_history(est_str):
    _, _, err, _, _, _, _, _, _, _ = pickle.load(open('/net/esker/archive1/pog/mldata/regressors/' +
                                                      est_str, + 'pkl', 'rb'))
    return err


